from ..querytypes.const import (
    QUERYTYPE_AVERAGEOFTHREEDAYS,
    QUERYTYPE_AVERAGEOFTHREEDAYS_MIN,
    QUERYTYPE_AVERAGEOFTHREEHOURS_MIN,
    QUERYTYPE_AVERAGEOFTHREEHOURS,
    QUERYTYPE_BASICMAX,
)
from ....models.locale.enums.querytype import QueryType
from ..querytypes.querytypes import QUERYTYPES
from ..locale_model import Locale_Type
from dataclasses import dataclass


@dataclass
class NO_Tensio(Locale_Type):
    observed_peak = QueryType.AverageOfThreeDays
    charged_peak = QueryType.AverageOfThreeDays
    query_model = QUERYTYPES[QueryType.AverageOfThreeDays]


@dataclass
class NO_LNett(Locale_Type):
    observed_peak = QueryType.AverageOfThreeHours
    charged_peak = QueryType.AverageOfThreeHours
    query_model = QUERYTYPES[QueryType.AverageOfThreeHours]


# docs: https://www.l-nett.no/nynettleie/slik-blir-ny-nettleie-og-pris


@dataclass
class NO_GlitreEnergi(Locale_Type):
    observed_peak = QueryType.Max
    charged_peak = QueryType.Max
    query_model = QUERYTYPES[QueryType.Max]


# docs: https://www.glitreenergi-nett.no/smart-nettleie/


@dataclass
class NO_AgderEnergi(Locale_Type):
    observed_peak = QueryType.AverageOfThreeDays
    charged_peak = QueryType.AverageOfThreeDays
    query_model = QUERYTYPES[QueryType.AverageOfThreeDays]


@dataclass
class NO_Elvia(Locale_Type):
    observed_peak = QueryType.AverageOfThreeDays
    charged_peak = QueryType.AverageOfThreeDays
    query_model = QUERYTYPES[QueryType.AverageOfThreeDays]


@dataclass
class NO_Lede(Locale_Type):
    observed_peak = QueryType.AverageOfThreeDays
    charged_peak = QueryType.AverageOfThreeDays
    query_model = QUERYTYPES[QueryType.AverageOfThreeDays]


@dataclass
class NO_BKK(Locale_Type):
    observed_peak = QueryType.AverageOfThreeDays
    charged_peak = QueryType.AverageOfThreeDays
    query_model = QUERYTYPES[QueryType.AverageOfThreeDays]


@dataclass
class NO_Mellom(Locale_Type):
    observed_peak = QueryType.AverageOfThreeDays
    charged_peak = QueryType.AverageOfThreeDays
    query_model = QUERYTYPES[QueryType.AverageOfThreeDays]


@dataclass
class NO_AskerNett(Locale_Type):
    observed_peak = QueryType.AverageOfThreeDays
    charged_peak = QueryType.AverageOfThreeDays
    query_model = QUERYTYPES[QueryType.AverageOfThreeDays]


# docs: https://askernett.no/ny-nettleiemodell-for-alle-fra-1-juli-2022/



from dataclasses import dataclass, field
from typing import Tuple

@dataclass
class MaxMinModel:
    input_hours: dict[int, Tuple[float, float]] = field(default_factory=lambda: {})
    original_input_hours: dict[int, Tuple[float, float]] = field(default_factory=lambda: {})
    expected_hourly_charge: float = 0

    @property
    def average_price(self) -> float|None:
        return self._caluclate_average_price(self.input_hours, self.total_charge, self.expected_hourly_charge)
    
    @property
    def original_average_price(self) -> float|None:
        return self._caluclate_average_price(self.original_input_hours, self.original_total_charge, self.expected_hourly_charge)

    @property
    def total_charge(self) -> float:
        return self._calculate_total_charge(self.input_hours, self.expected_hourly_charge)
    
    @property
    def original_total_charge(self) -> float:
        return self._calculate_total_charge(self.original_input_hours, self.expected_hourly_charge)


    @staticmethod
    def _caluclate_average_price(input:dict, total_charge: float, expected: float) -> float|None:
        try:
            first = sum([v[0] * v[1] * expected for k,v in input.items() if v[1] > 0])
            return round(first/total_charge, 2)
        except ZeroDivisionError as e:
            return None

    @staticmethod
    def _calculate_total_charge(input:dict, expected:float) -> float:
        ret = sum([v[1] * expected for k,v in input.items()])
        return round(ret, 1)



from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ..hoursselection import Hoursselection
from typing import Tuple
from ....models.hourselection.max_min_model import MaxMinModel

MINIMUM_DIFFERENCE = 0.05

class MaxMinCharge:
    def __init__(self, hoursselection: Hoursselection) -> None:
        self.model = MaxMinModel()
        self.parent = hoursselection
        self.active: bool = False

    @property
    def total_charge(self) -> float:
        return self.model.total_charge
    
    @property
    def average_price(self) -> float|None:
        return self.model.average_price
    
    @property
    def original_total_charge(self) -> float:
        return self.model.original_total_charge
    
    @property
    def original_average_price(self) -> float|None:
        return self.model.original_average_price


    @property
    def non_hours(self) -> list:
        return [k for k, v in self.model.input_hours.items() if v[1] == 0]
    
    @property
    def dynamic_caution_hours(self) -> dict:
        return {k: v[1] for k, v in self.model.input_hours.items() if 0 < v[1] < 1}

    async def async_update(self, avg24, peak, max_desired) -> None:
        _avg24 = round((avg24/1000),1)
        self.model.expected_hourly_charge = peak-_avg24
        for i in range(len(self.model.original_input_hours.items())):
            if self.total_charge - max_desired > MINIMUM_DIFFERENCE:
                await self.async_decrease()
            elif self.total_charge - max_desired < MINIMUM_DIFFERENCE*-1:
                expected_charge = round((max_desired-self.total_charge)/(peak-_avg24),2)
                await self.async_increase(expected_charge)
            if abs(self.total_charge - max_desired) < MINIMUM_DIFFERENCE:
                break        

    async def async_initial_charge(self, avg24, peak) -> float:
        _avg24 = round((avg24/1000),1)
        self.model.expected_hourly_charge = peak-_avg24
        total=24*(peak-_avg24) #todo: fix 24 to be dynamic
        total -= len(self.non_hours)*(peak-_avg24)
        total -= sum(self.dynamic_caution_hours.values())*(peak-_avg24)
        return total

    async def async_sum_charge(self, avg24, peak) -> float:
        total=0
        for k, v in self.model.input_hours.items():
            total += (peak-avg24)*v[1]
        return total

    async def async_decrease(self):
        max_key = max(
            self.model.input_hours, 
            key=lambda k: self.model.input_hours[k][0] if self.model.input_hours[k][1] != 0 else -1
            )
        self.model.input_hours[max_key] = (self.model.input_hours[max_key][0], 0)

    async def async_increase(self, expected_charge):
        min_key = min(
            self.model.input_hours, 
            key=lambda k: self.model.input_hours[k][0] if self.model.input_hours[k][1] < 1 and self.model.original_input_hours[k][1] > 0 else 999
            )
        self.model.input_hours[min_key] = (self.model.input_hours[min_key][0], min(
            min(1, expected_charge), 
            self.model.original_input_hours[min_key][1]
            ))

    async def async_setup(
            self, 
            max_charge: float,
            non_hours:list|None = None, 
            dynamic_caution_hours:dict|None = None, 
            prices:list|None = None, 
            prices_tomorrow:list|None = None,
            mock_hour: int|None = None
            ) -> None:
        
        if max_charge == 0:
            self.active = False
            return        
        hour = mock_hour if mock_hour is not None else await self.parent.service.async_set_hour()
        _non_hours = self.parent.internal_non_hours if non_hours is None else non_hours
        _dynamic_caution_hours = self.parent.internal_dynamic_caution_hours if dynamic_caution_hours is None else dynamic_caution_hours
        _prices = self.parent.prices if prices is None else prices
        _prices_tomorrow = self.parent.prices_tomorrow if prices_tomorrow is None else prices_tomorrow
        ret_today, ret_tomorrow = await self.async_loop_nonhours(hour, _non_hours, _prices, _prices_tomorrow)
        await self.async_loop_caution_hours(hour, _dynamic_caution_hours, _prices, _prices_tomorrow, ret_today, ret_tomorrow)        
        await self.async_add_available_hours(hour, _prices, _prices_tomorrow, ret_today, ret_tomorrow)
        self.model.input_hours = await self.async_sort_dicts(ret_today, ret_tomorrow)        
        self.model.original_input_hours = self.model.input_hours.copy()
        self.active = True

    @staticmethod
    async def async_add_available_hours(hour: int, prices:list, prices_tomorrow:list, ret_today: dict, ret_tomorrow:dict) -> None:
        _hour = hour
        _range = 24 if len(prices_tomorrow) > 0 else len(prices) - hour
        for i in range(_range):
            if _hour < hour and _hour not in ret_tomorrow.keys():
                ret_tomorrow[_hour] = (prices_tomorrow[_hour], 1)
                #todo: must test this without valid prices tomorrow.
            elif _hour >= hour and _hour not in ret_today.keys():
                ret_today[_hour] = (prices[_hour], 1)
            _hour += 1
            if _hour > 23:
                _hour = 0

    @staticmethod
    async def async_loop_caution_hours(hour: int, caution_hours:dict, prices:list, prices_tomorrow:list, ret_today: dict, ret_tomorrow: dict) -> None:
        for k, v in caution_hours.items():
            if k >= hour:
                ret_today[k] = (prices[k], v)
            elif len(prices_tomorrow) > 0:
                ret_tomorrow[k] = (prices_tomorrow[k], v)

    @staticmethod
    async def async_loop_nonhours(hour: int, non_hours:list, prices:list, prices_tomorrow:list) -> Tuple[dict, dict]:
        ret_today = {}
        ret_tomorrow = {}
        for n in non_hours:
            if n >= hour:
                ret_today[n] = (prices[n], 0)
            elif len(prices_tomorrow) > 0:
                ret_tomorrow[n] = (prices_tomorrow[n], 0)
        return ret_today, ret_tomorrow

    @staticmethod
    async def async_sort_dicts(ret_today: dict, ret_tomorrow:dict) -> dict:
        ret = {}
        for k in sorted(ret_today.keys()):
            ret[k] = ret_today[k]
        for k in sorted(ret_tomorrow.keys()):
            ret[k] = ret_tomorrow[k]
        return ret






from .session_service import SessionService
import logging

_LOGGER = logging.getLogger(__name__)


class Session:
    def __init__(self, charger):
        self._charger = charger
        self.service: SessionService = SessionService()

    async def async_setup(self):
        await self.service.async_setup()

    @property
    def session_energy(self):
        return self.service.total_energy

    async def async_set_session_energy(self, val):
        await self.service.async_update_power_reading(val)
        await self.async_update_session_pricing()

    @property
    def session_price(self):
        return self.service.total_price

    async def async_set_session_price(self, val) -> None:
        await self.service.async_update_price(val)
        await self.async_update_session_pricing()

    @property
    def energy_average(self) -> float:
        return self.service.energy_average

    @property
    def energy_weekly_dict(self) -> dict:
        return self.service.energy_weekly_dict

    async def async_update_session_pricing(self):
        if self._charger.model.session_active is False:
            await self.async_terminate()

    async def async_terminate(self):
        await self.service.async_terminate()

    async def async_reset(self):
        await self.service.async_reset()

    async def async_unpack(self, data):
        await self.service.average_data.async_unpack(data)

    async def async_setup_fresh(self):
        await self.service.average_data.async_set_init_model()



from datetime import datetime
from ...models.chargecontroller_states import ChargeControllerStates
from .scheduler import Scheduler


class SchedulerFacade(Scheduler):
    def __init__(self, hub, options):
        self.hub = hub
        super().__init__(options)
        self.schedule_created = False

    async def async_create_schedule(self, charge_amount: float, departure_time: datetime, schedule_starttime: datetime, override_settings: bool = False):
        if not self.scheduler_active:
            await self.async_create(charge_amount, departure_time, schedule_starttime, override_settings)
        self.schedule_created = True

    async def async_update_facade(self):
        await self.async_update(self.hub.sensors.powersensormovingaverage24.value,
                                self.hub.current_peak_dynamic,
                                self.hub.chargecontroller.charger.session.session_energy,
                                self.hub.hours.prices,
                                self.hub.hours.prices_tomorrow)
        await self.async_check_states()

    async def async_cancel_facade(self):
        await self.async_cancel()
        self.schedule_created = False

    async def async_check_states(self):
        if not self.scheduler_active and self.schedule_created:
            await self.async_cancel_facade()
        elif self.hub.chargecontroller.status_string is ChargeControllerStates.Done.name:
            await self.async_cancel_facade()

    @property
    def non_hours(self) -> list:
        return self.model.non_hours

    @property
    def caution_hours(self) -> dict:
        """dynamic caution hours"""
        return self.model.caution_hours


import logging
from enum import Enum

_LOGGER = logging.getLogger(__name__)


class Fuses(Enum):
    FUSE_3_16 = "3phase 16A"
    FUSE_3_20 = "3phase 20A"
    FUSE_3_25 = "3phase 25A"
    FUSE_3_35 = "3phase 35A"
    FUSE_3_50 = "3phase 50A"
    FUSE_3_63 = "3phase 63A"
    DEFAULT = "Not set"

    def parse_from_config(fusetype: str):
        try:
            for f in Fuses:
                if fusetype == f.value:
                    return f
        except Exception as e:
            _LOGGER.error("Unable to parse fuse-type, invalid value: {e}")
            return Fuses.DEFAULT


from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .locale_model import Locale_Type

from .countries.default import Default, NoPeak
from .countries.belgium import VregBelgium
from .countries.sweden import (
    SE_Bjerke_Energi,
    SE_FalbygdensEnergi,
    SE_Gothenburg,
    SE_Karlstad,
    SE_Kristinehamn,
    SE_Linde_Energi,
    SE_Malarenergi,
    SE_Malung_Salen,
    SE_Nacka_normal,
    SE_NACKA_timediff,
    SE_Partille,
    SE_SHE_AB,
    SE_Skovde,
    SE_Sollentuna,
    SE_TekniskaVerken_Link,
    SE_Eskilstuna,
    SE_Ellevio,
    SE_JBF,
    # SE_Telge_Energi
)
from .countries.norway import (
    NO_AgderEnergi,
    NO_Elvia,
    NO_GlitreEnergi,
    NO_LNett,
    NO_Lede,
    NO_Mellom,
    NO_Tensio,
    NO_BKK,
    NO_AskerNett,
)


"""LOCALETYPES"""
LOCALE_SE_ESKILSTUNA = "Eskilstuna elnät, Sweden"
LOCALE_SE_GOTHENBURG = "Gothenburg, Sweden"
LOCALE_SE_KARLSTAD = "Karlstad, Sweden"
LOCALE_SE_KRISTINEHAMN = "Kristinehamn, Sweden"
LOCALE_SE_NACKA_NORMAL = "Nacka, Sweden (Normal tariffe)"
# LOCALE_SE_NACKA_TIMEDIFF = "Nacka, Sweden (Time differentiated tariffe)"
LOCALE_SE_PARTILLE = "Partille, Sweden"
LOCALE_DEFAULT = "Other, just want to test"
LOCALE_NO_PEAK = "No peak shaving needed"
LOCALE_SE_SALA = "Sala-Heby Energi AB, Sweden"
LOCALE_SE_MALUNG_SALEN = "Malung-Sälen, Sweden (Malungs elverk)"
LOCALE_SE_TEKNISKA_LINK = "Tekniska verken Linköping, Sweden"
LOCALE_SE_SKOVDE = "Skövde, Sweden"
LOCALE_SE_SOLLENTUNA = "Sollentuna Energi, Sweden"
LOCALE_BE_VREG = "Belgium (VREG)"
LOCALE_SE_BJERKE_ENERGI = "Bjärke Energi, Sweden"
LOCALE_NO_GLITRE_ENERGI = "Glitre Energi, Norway"
LOCALE_NO_AGDER_ENERGI = "Agder Energi, Norway"
LOCALE_NO_LNETT = "LNett, Norway"
LOCALE_NO_TENSIO = "Tensio, Norway"
LOCALE_SE_LINDE_ENERGI = "Linde Energi, Sweden"
LOCALE_SE_FALBYDGENS_ENERGI = "Falbygdens Energi, Sweden"
LOCALE_NO_ELVIA = "Elvia, Norway"
LOCALE_NO_LEDE = "Lede, Norway"
LOCALE_NO_BKK = "BKK, Norway"
LOCALE_SE_MALARENERGI = "Mälarenergi, Sweden"
LOCALE_NO_MELLOM = "Mellom, Norway"
LOCALE_NO_ASKER = "Asker Nett, Norway"
LOCALE_SE_ELLEVIO = "Ellevio, Sweden"
LOCALE_SE_JBF = "Jukkasjärvi (JBF), Sweden"

LOCALETYPEDICT = {
    LOCALE_DEFAULT: Default,
    LOCALE_NO_PEAK: NoPeak,
    LOCALE_SE_ESKILSTUNA: SE_Eskilstuna,
    LOCALE_SE_GOTHENBURG: SE_Gothenburg,
    LOCALE_SE_PARTILLE: SE_Partille,
    LOCALE_SE_KARLSTAD: SE_Karlstad,
    LOCALE_SE_KRISTINEHAMN: SE_Kristinehamn,
    LOCALE_SE_NACKA_NORMAL: SE_Nacka_normal,
    LOCALE_SE_MALUNG_SALEN: SE_Malung_Salen,
    LOCALE_SE_SALA: SE_SHE_AB,
    LOCALE_SE_SKOVDE: SE_Skovde,
    LOCALE_SE_SOLLENTUNA: SE_Sollentuna,
    LOCALE_SE_TEKNISKA_LINK: SE_TekniskaVerken_Link,
    LOCALE_BE_VREG: VregBelgium,
    LOCALE_SE_BJERKE_ENERGI: SE_Bjerke_Energi,
    LOCALE_NO_GLITRE_ENERGI: NO_GlitreEnergi,
    LOCALE_NO_AGDER_ENERGI: NO_AgderEnergi,
    LOCALE_NO_LNETT: NO_LNett,
    LOCALE_NO_TENSIO: NO_Tensio,
    LOCALE_SE_LINDE_ENERGI: SE_Linde_Energi,
    LOCALE_SE_FALBYDGENS_ENERGI: SE_FalbygdensEnergi,
    LOCALE_NO_ELVIA: NO_Elvia,
    LOCALE_NO_LEDE: NO_Lede,
    LOCALE_NO_BKK: NO_BKK,
    LOCALE_SE_MALARENERGI: SE_Malarenergi,
    LOCALE_NO_MELLOM: NO_Mellom,
    LOCALE_NO_ASKER: NO_AskerNett,
    LOCALE_SE_ELLEVIO: SE_Ellevio,
    LOCALE_SE_JBF: SE_JBF,
}

"""Lookup locales for config flow"""
LOCALES = [
    LOCALE_BE_VREG,
    LOCALE_NO_AGDER_ENERGI,
    LOCALE_NO_BKK,
    LOCALE_NO_ELVIA,
    LOCALE_NO_GLITRE_ENERGI,
    LOCALE_NO_LEDE,
    LOCALE_NO_LNETT,
    LOCALE_NO_MELLOM,
    LOCALE_NO_TENSIO,
    LOCALE_NO_ASKER,
    LOCALE_SE_BJERKE_ENERGI,
    LOCALE_SE_ELLEVIO,
    LOCALE_SE_FALBYDGENS_ENERGI,
    LOCALE_SE_GOTHENBURG,
    LOCALE_SE_JBF,
    LOCALE_SE_KRISTINEHAMN,
    LOCALE_SE_LINDE_ENERGI,
    LOCALE_SE_MALUNG_SALEN,
    LOCALE_SE_MALARENERGI,
    LOCALE_SE_NACKA_NORMAL,
    LOCALE_SE_PARTILLE,
    LOCALE_SE_SALA,
    LOCALE_SE_SKOVDE,
    LOCALE_SE_SOLLENTUNA,
    LOCALE_SE_TEKNISKA_LINK,
    LOCALE_DEFAULT,
    LOCALE_NO_PEAK,
]


class LocaleData:
    def __init__(self, input_type: str, domain: str):
        self._type = input_type
        self._domain = domain
        self._data: Locale_Type = LOCALETYPEDICT[input_type]()

    @property
    def type(self) -> str:
        return self._type

    @property
    def data(self) -> Locale_Type:
        return self._data

    async def async_try_update_peak(self, new_val, timestamp):
        """Try to update peak data."""
        if self._data is not None:
            await self._data.query_model.async_try_update(new_val, timestamp)


class LocaleFactory:
    @staticmethod
    async def async_create(input_type: str, domain: str = "peaqev"):
        """Create a locale object."""
        ret = LocaleData(input_type, domain)
        await ret.data.async_set_query_service()
        return ret



from dataclasses import dataclass, field


@dataclass
class SessionModel:
    total_price: float = 0
    price: float = 0
    current_power: float = 0
    total_energy: float = 0
    current_time: float = 0
    time_delta: float = 0
    readings: list = field(default_factory=lambda: [])



from dataclasses import dataclass, field
from .savings_status import SavingsStatus
from datetime import date, datetime
from typing import Tuple
import logging

_LOGGER = logging.getLogger(__name__)


@dataclass
class SavingsModel:
    peak_price_per_kwh: float
    car_connected_at: datetime | None = None
    prices: dict[date, list[float]] = field(default_factory=lambda: {})
    consumption: dict[date, dict[int, float]] = field(default_factory=lambda: {})
    # peaks: dict[date, dict[int, float]] = field(default_factory=lambda: {})
    status: SavingsStatus = SavingsStatus.Off

    async def async_reset(self) -> None:
        self.status = SavingsStatus.Off
        self.prices = {}
        self.consumption = {}
        # self.peaks = {}

    async def async_re_initialize(self, incoming: dict):
        self.car_connected_at = incoming.get("car_connected_at", None)
        self.prices = incoming.get("prices", {})
        self.consumption = incoming.get("consumption", {})
        # self.peaks = incoming.get("peaks", {})
        if all(
            [
                self.car_connected_at is not None,
                len(self.prices) > 0,
                len(self.consumption) > 0,
                # len(self.peaks) > 0,
            ]
        ):
            self.status = SavingsStatus.Collecting
        else:
            _LOGGER.warning("Cannot re-initialize SavingsModel, missing data")

    async def async_add_prices(
        self, prices: list[float], _date: date | None = None
    ) -> None:
        if _date is None:
            _date = datetime.now().date()
        self.prices[_date] = prices

    async def async_add_to_consumption(
        self,
        consumption: float,
        date: date | None = None,
        hour: int | None = None,
    ) -> None:
        if self.status is not SavingsStatus.Collecting:
            return
        _date, _hour = await self.async_check_date_hour(date, hour)
        await self.async_try_add(_date, _hour, self.consumption)
        self.consumption[_date][_hour] = max(
            [consumption, self.consumption[_date][_hour]]
        )

    # async def async_add_to_peaks(
    #     self, peak: float, date: date | None = None, hour: int | None = None
    # ) -> None:
    #     if self.status is not SavingsStatus.Collecting:
    #         return
    #     _date, _hour = await self.async_check_date_hour(date, hour)
    #     await self.async_try_add(_date, _hour, self.peaks)
    #     self.peaks[_date][_hour] = max([peak, self.peaks[_date][_hour]])

    async def async_check_date_hour(
        self, _date: date | None, _hour: int | None
    ) -> Tuple[date, int]:
        if _date is None:
            _date = date.today()
        if _hour is None:
            _hour = datetime.now().hour
        return _date, _hour

    async def async_try_add(self, _date, _hour, modeldict: dict):
        if _date not in modeldict:
            modeldict[_date] = {}
        if _hour not in modeldict[_date].keys():
            modeldict[_date][_hour] = 0



from dataclasses import dataclass

@dataclass
class ServiceCallsOptions:
    allowupdatecurrent: bool
    update_current_on_termination: bool
    switch_controls_charger: bool = False
    """
    This is for outlet control. 
    If false we call services to enable/disable. 
    If true we call a switch
    """


from datetime import datetime
from ...models.chargecontroller_states import ChargeControllerStates
from .thresholdbase import ThresholdBase


class Threshold(ThresholdBase):
    def __init__(self, hub):
        self._hub = hub
        super().__init__(hub)

    @property
    def allowedcurrent(self) -> int:
        amps = self._setcurrentdict()
        if (
            self._hub.chargecontroller.status_string
            is not ChargeControllerStates.Start.name
        ):
            return min(amps.values())
        return ThresholdBase.allowed_current(
            datetime.now().minute,
            self._hub.sensors.powersensormovingaverage.value
            if self._hub.sensors.powersensormovingaverage.value is not None
            else 0,
            self._hub.sensors.charger_enabled.value,
            self._hub.sensors.charger_done.value,
            amps,
            self._hub.sensors.totalhourlyenergy.value,
            self._hub.current_peak_dynamic,
            self._hub.sensors.locale.data.is_quarterly(),
            self._hub.power.power_canary.max_current_amp,
        )

    async def async_allowed_current(self) -> int:
        amps = await self.async_setcurrentdict()
        if (
            self._hub.chargecontroller.status_string
            is not ChargeControllerStates.Start.name
        ):
            return min(amps.values())
        return await ThresholdBase.async_allowed_current(
            datetime.now().minute,
            self._hub.sensors.powersensormovingaverage.value
            if self._hub.sensors.powersensormovingaverage.value is not None
            else 0,
            self._hub.sensors.charger_enabled.value,
            self._hub.sensors.charger_done.value,
            amps,
            self._hub.sensors.totalhourlyenergy.value,
            self._hub.current_peak_dynamic,
            self._hub.sensors.locale.data.is_quarterly(),
            self._hub.power.power_canary.max_current_amp,
        )



from ....models.locale.enums.time_periods import TimePeriods
from ....models.locale.enums.querytype import QueryType
from ..querytypes.querytypes import QUERYTYPES
from dataclasses import dataclass
from ..locale_model import Locale_Type


@dataclass
class VregBelgium(Locale_Type):
    observed_peak = QueryType.Max
    charged_peak = QueryType.Max
    query_model = QUERYTYPES[QueryType.Max]
    # TODO: to be decided. Should charged_peak be turned into the real charged peak, ie the average of the months in a year? could be issues with the long term stats there and it won't help peaq in any way.
    peak_cycle = TimePeriods.QuarterHourly


# https://www.vreg.be/nl/nieuwe-nettarieven



from datetime import datetime
from typing import List
from ....models.locale.enums.dividents import Dividents
from ....models.locale.enums.datepart_modeltype import DatePartModelType
from ....models.locale.enums.datepart_datetype import DatePartDateType
from ....models.locale.queryservice_model import QueryServiceModel


class QueryService:
    def __init__(self, args: QueryServiceModel = QueryServiceModel()):
        self._settings = args

    # def should_register_peak(self, dt: datetime) -> bool:
    #     main_ret = []
    #     main_grouping = (
    #         s for s in self._settings.groups if s.divident is not Dividents.UNSET
    #     )
    #     for s in main_grouping:
    #         group_ret = []
    #         grouping = (a for a in s.dateparts if len(a.values))
    #         for a in grouping:
    #             group_ret.append(QueryService.datepart(a.type, a.dttype, a.values, dt))
    #         if s.divident is Dividents.AND:
    #             main_ret.append(all(group_ret))
    #         else:
    #             main_ret.append(any(group_ret))
    #     return any(main_ret) if len(main_ret) else True

    # @staticmethod
    # def datepart(
    #     logic: str, dtpart: str, arguments: List[int], timer: datetime
    # ) -> bool:
    #     if not arguments:
    #         return True
    #     arg = arguments if len(arguments) > 1 else arguments[0]
    #     _logic = QueryService.LOGIC.get(logic)(
    #         QueryService.DATETIMEPARTS.get(dtpart)(timer), arg
    #     )
    #     return _logic

    async def async_should_register_peak(self, dt: datetime) -> bool:
        main_ret = []
        for s in [
            s for s in self._settings.groups if s.divident is not Dividents.UNSET
        ]:
            group_ret = []
            grouping = (a for a in s.dateparts if len(a.values))
            for a in grouping:
                group_ret.append(
                    await QueryService.async_datepart(a.type, a.dttype, a.values, dt)
                )
            if s.divident is Dividents.AND:
                main_ret.append(all(group_ret))
            else:
                main_ret.append(any(group_ret))
        print(main_ret)
        return any(main_ret) if len(main_ret) else True

    @staticmethod
    async def async_datepart(
        logic: str, dtpart: str, arguments: List[int], timer: datetime
    ) -> bool:
        if not arguments:
            return True
        try:
            arg = arguments if len(arguments) > 1 else arguments[0]
            _logic = QueryService.LOGIC.get(logic)(
                QueryService.DATETIMEPARTS.get(dtpart)(timer), arg
            )
            return _logic
        except Exception as e:
            raise Exception(f"Error in async_datepart: {e}")

    AND = "AND"
    OR = "OR"
    LOGIC = {
        DatePartModelType.Equal: lambda a, dtp: dtp == a,
        DatePartModelType.Less: lambda a, dtp: a < dtp,
        DatePartModelType.Greater: lambda a, dtp: a > dtp,
        DatePartModelType.Not: lambda a, dtp: dtp != a,
        DatePartModelType.LessOrEqual: lambda a, dtp: a <= dtp,
        DatePartModelType.GreaterOrEqual: lambda a, dtp: a >= dtp,
        DatePartModelType.In: lambda a, dtp: a in dtp,
    }
    DATETIMEPARTS = {
        DatePartDateType.Weekday: lambda d: d.weekday(),
        DatePartDateType.Month: lambda d: d.month,
        DatePartDateType.Hour: lambda d: d.hour,
    }



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ...services.locale.querytypes.queryservice import QueryService
from dataclasses import dataclass, field
from datetime import datetime
from ...models.locale.enums.querytype import QueryType
from ...models.locale.enums.time_periods import TimePeriods
from ..locale.time_pattern import TimePattern
from ...models.locale.price.locale_price import LocalePrice
from .querytypes.querytypes import LocaleQuery


@dataclass
class Locale_Type:
    observed_peak: QueryType = field(init=False)
    charged_peak: QueryType = field(init=False)
    query_model: LocaleQuery = field(init=False)
    query_service: QueryService | None = None
    price: LocalePrice = LocalePrice()
    free_charge_pattern: TimePattern | None = None
    peak_cycle: TimePeriods = TimePeriods.Hourly

    def free_charge(self, mockdt: datetime = datetime.min) -> bool:
        if self.free_charge_pattern is None:
            return False
        now = datetime.now() if mockdt is datetime.min else mockdt
        return self.free_charge_pattern.valid(now)

    def is_quarterly(self) -> bool:
        return self.peak_cycle == TimePeriods.QuarterHourly

    async def async_free_charge(self, mockdt: datetime = datetime.min) -> bool:
        if self.free_charge_pattern is None:
            return False
        now = datetime.now() if mockdt is datetime.min else mockdt
        return await self.free_charge_pattern.async_valid(now)

    async def async_is_quarterly(self) -> bool:
        return self.peak_cycle == TimePeriods.QuarterHourly

    async def async_set_query_service(self):
        if self.query_service is not None:
            await self.query_model.async_set_query_service(self.query_service)



import setuptools

setuptools.setup(
    name="peaqevcore",
    version="16.0.3",
    author="Magnus Eldén",
    description="Core types for peaqev car charging",
    url="https://github.com/elden1337/peaqev-core",
    license="CC-NC-ND",
    packages=[
        "peaqevcore",
        "peaqevcore.services.hourselection",
        "peaqevcore.services.hourselection.hourselectionservice",
        "peaqevcore.services.hourselection.initializers",
        "peaqevcore.models.hourselection",
        "peaqevcore.models.hourselection.hourobjects",
        "peaqevcore.services.chargertype",
        "peaqevcore.models.chargertype",
        "peaqevcore.models",
        "peaqevcore.models.hub",
        "peaqevcore.hub",
        "peaqevcore.services.prediction",
        "peaqevcore.services.savings",
        "peaqevcore.services.session",
        "peaqevcore.services.threshold",
        "peaqevcore.services.locale",
        "peaqevcore.services.locale.countries",
        "peaqevcore.services.locale.querytypes",
        "peaqevcore.models.locale",
        "peaqevcore.models.locale.enums",
        "peaqevcore.models.locale.price",
        "peaqevcore.models.locale.price.models",
        "peaqevcore.services.scheduler",
        "peaqevcore.services.timer",
    ],
    extras_require={"tests": ["pytest>3.6.4", "pytest-cov<2.6", "pytest-asyncio"]},
)



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..hoursselection import Hoursselection

import logging
from datetime import datetime
from typing import Tuple
import statistics as stat

from .hoursselection_helpers import async_create_dict
from .hourselection_calculations import async_normalize_prices, async_create_cautions, async_get_offset_dict, async_should_be_cautionhour, async_set_charge_allowance
from ....models.hourselection.hourobjects.hourobject import HourObject
from ....models.hourselection.hourobjects.hourobject_helpers import async_update_interim_lists
from ....models.hourselection.day_types import DayTypes

_LOGGER = logging.getLogger(__name__)


class HourSelectionService:
    def __init__(self, parent: Hoursselection, base_mock_hour: int|None = None):
        self.parent = parent
        self._mock_hour = base_mock_hour
        self._mock_day:int|None = None
        self.preserve_interim: bool = False
        self._midnight_touched: bool = False

    async def async_update(self) -> None:
        if all([len(self.parent.model.prices_today) == 0, len(self.parent.model.prices_tomorrow) == 0]):
            return
        if self.preserve_interim: 
            await self.async_change_midnight_data()  
        else:
            self._midnight_touched = False
            today= await self.async_update_per_day(prices=self.parent.model.prices_today, day_type=DayTypes.Today)
            tomorrow= await self.async_update_per_day(prices=self.parent.model.prices_tomorrow, day_type=DayTypes.Tomorrow)
            hours, hours_tomorrow = await self.async_interim_day_update(today, tomorrow)

            self.parent.model.hours.hours_today = await self.async_add_remove_limited_hours(hours, day_type=DayTypes.Today)
            self.parent.model.hours.hours_tomorrow = await self.async_add_remove_limited_hours(hours_tomorrow, day_type=DayTypes.Tomorrow)
            hour = await self.async_set_hour()
            await self.parent.model.hours.async_update_hour_lists(hour=hour)

    async def async_change_midnight_data(self) -> None:        
        if self.parent.model.prices_tomorrow == []:
            if not self._midnight_touched:
                self._midnight_touched = await self.parent.model.hours.async_touch_midnight()
        else:
            self.preserve_interim = False
            await self.async_update()

    async def async_update_per_day(self, prices: list, day_type: DayTypes, range_start: int =0) -> HourObject:
        _LOGGER.debug(f"Updating {day_type.name}")
        pricedict = {}
        if prices is not None and len(prices) > 1:
            pricedict = await async_create_dict(prices)
            normalized_pricedict = await async_create_dict(
                await async_normalize_prices(prices)
                )
            if stat.stdev(prices) > 0.05:
                ranked = await async_create_cautions(
                        pricedict, 
                        normalized_pricedict,
                        self.parent.cautionhour_type_enum,
                        range_start,
                        self.parent.model.adjusted_average,
                        self.parent.model.options.blocknocturnal
                        )
                ready_hours = await self.async_determine_hours(ranked, prices)
                ret= HourObject(
                    nh=ready_hours.nh, 
                    ch=ready_hours.ch, 
                    dyn_ch=ready_hours.dyn_ch, 
                    pricedict=pricedict
                    )
            else:
                """the price curve is too flat to determine hours"""
                ret= HourObject(nh=[], ch=[], dyn_ch={},pricedict=pricedict)
            ret.offset_dict= await async_get_offset_dict(normalized_pricedict)
            return ret
        return HourObject([],[],{})

    async def async_add_remove_limited_hours(self, hours: HourObject, day_type:DayTypes) -> HourObject:
        """Removes cheap hours and adds expensive hours set by user limitation"""
        if hours is None or all(
            [
                len(hours.nh) == 0, 
                len(hours.ch) == 0, 
                len(hours.dyn_ch) == 0
            ]):
            return HourObject(
                offset_dict=hours.offset_dict,
                pricedict=hours.pricedict
                )
        match day_type:
            case DayTypes.Today | DayTypes.Interim:
                await hours.async_add_expensive_hours(self.parent.model.options.absolute_top_price)
            case DayTypes.Tomorrow:
                _top = await self.parent.model.options.async_add_tomorrow_to_top_price(
                    self.parent.prices_tomorrow, 
                    self._mock_day
                    )
                await hours.async_add_expensive_hours(_top)     
        return await hours.async_remove_cheap_hours(self.parent.model.options.min_price)

    async def async_determine_hours(self, price_list: dict, prices: list) -> HourObject:
        ret = HourObject([],[],{})
        try:
            for p in price_list:    
                if price_list[p]["force_non"] is True:
                    ret.nh.append(p)
                elif await async_should_be_cautionhour(
                    price_list[p], 
                    prices, 
                    self.parent.model.current_peak, 
                    self.parent.model.options.cautionhour_type
                    ):
                    ret.ch.append(p)
                    ret.dyn_ch[p] = round(
                        await async_set_charge_allowance(
                            price_list[p]["permax"], 
                            self.parent.model.options.cautionhour_type
                            ),
                        2)
                else:
                    ret.nh.append(p)
        except IndexError as e:
            _LOGGER.error(f"Error in determine hours: {e}")
        return ret
    
    def set_hour(self, testhour:int|None = None) -> int:
        return testhour if testhour is not None else self._mock_hour if self._mock_hour is not None else datetime.now().hour
    
    async def async_set_hour(self, testhour:int|None = None) -> int:
        return testhour if testhour is not None else self._mock_hour if self._mock_hour is not None else datetime.now().hour
    
    async def async_set_day(self, day:int|None = None):
        self._mock_day = day if day is not None else datetime.now().day

    async def async_interim_day_update(self, today: HourObject, tomorrow: HourObject) -> Tuple[HourObject, HourObject]:
        """Updates the non- and caution-hours with an adjusted mean of 14h - 13h today-tomorrow to get a more sane nightly curve."""
        if len(self.parent.model.prices_tomorrow) < 23:
            self.preserve_interim = False
            return today, tomorrow 
        
        hour =len(self.parent.prices)-10
        negative_hour = (len(self.parent.prices) - hour)*-1
        pricelist = self.parent.model.prices_today[hour::]
        pricelist[len(pricelist):] = self.parent.model.prices_tomorrow[0:hour]
        new_hours = await self.async_update_per_day(prices=pricelist, range_start=hour, day_type=DayTypes.Interim)
        
        today = await async_update_interim_lists(range(hour,len(self.parent.prices)), today, new_hours, hour)
        tomorrow = await async_update_interim_lists(range(0,hour), tomorrow, new_hours, negative_hour)
        self.preserve_interim = True
        return today, tomorrow



from ....services.locale.querytypes.queryservice import QueryService
from ....models.locale.enums.querytype import QueryType
from ..time_pattern import TimePattern
from ....models.locale.enums.calendar_periods import CalendarPeriods
from ..locale_model import Locale_Type
from dataclasses import dataclass
from ..querytypes.querysets import QUERYSETS
from ..querytypes.querytypes import QUERYTYPES
from ..querytypes.const import QUERYTYPE_NEVER


@dataclass
class Default(Locale_Type):
    observed_peak = QueryType.Max
    charged_peak = QueryType.Max
    query_model = QUERYTYPES[QueryType.Max]


@dataclass
class NoPeak(Locale_Type):
    observed_peak = QueryType.Max
    charged_peak = QueryType.Max
    query_model = QUERYTYPES[QueryType.Max]
    query_service = QueryService(QUERYSETS[QUERYTYPE_NEVER])
    free_charge_pattern = TimePattern(
        [
            {
                CalendarPeriods.Month: [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
                CalendarPeriods.Weekday: [0, 1, 2, 3, 4, 5, 6],
                CalendarPeriods.Hour: [
                    0,
                    1,
                    2,
                    3,
                    4,
                    5,
                    6,
                    7,
                    8,
                    9,
                    10,
                    11,
                    12,
                    13,
                    14,
                    15,
                    16,
                    17,
                    18,
                    19,
                    20,
                    21,
                    22,
                    23,
                ],
            }
        ]
    )
